import java.awt.Dimension;
import java.io.File;
import java.net.URL;
import java.text.DecimalFormat;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import javax.swing.Box;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JCheckBox;
import javax.swing.JOptionPane;
import javax.swing.SwingConstants;

import com.treestar.flowjo.application.derivedparameters.DerivedParameterNode;
import com.treestar.flowjo.core.AlgebraicParameter;
import com.treestar.flowjo.core.DerivedParameter;
import com.treestar.flowjo.core.Sample;
import com.treestar.flowjo.core.histogram.DiscreteFunction;
import com.treestar.flowjo.core.histogram.PeakFinder;
import com.treestar.flowjo.core.histogram.PeakInfo;
import com.treestar.flowjo.core.histogram.PeakList;
import com.treestar.flowjo.engine.FEML;
import com.treestar.lib.FJPluginHelper;
import com.treestar.lib.PluginHelper;
import com.treestar.lib.core.ExportFileTypes;
import com.treestar.lib.core.ExternalAlgorithmResults;
import com.treestar.lib.core.ExternalPopulationAlgorithmInterface;
import com.treestar.lib.fcs.ParameterUtil;
import com.treestar.lib.fjml.FJML;
import com.treestar.lib.gui.GuiFactory;
import com.treestar.lib.gui.HBox;
import com.treestar.lib.gui.numberfields.IntlFormatter;
import com.treestar.lib.gui.numberfields.RangedDoubleTextField;
import com.treestar.lib.gui.numberfields.RangedIntegerTextField;
import com.treestar.lib.gui.panels.FJLabel;
import com.treestar.lib.gui.panels.FJPanel;
import com.treestar.lib.gui.swing.FJComboBox;
import com.treestar.lib.gui.text.FJTextField;
import com.treestar.lib.xml.SElement;

public class Calibration implements ExternalPopulationAlgorithmInterface
{
	// Variable section
	private int minimumGates 			= 0;
	private int histoSize 				= 4096;
	private int numPeaksFound 			= 0;
	private double[] linRegMean			= null;
	private double[] linRegAbs 			= null;
	private boolean checkAuto 			= false;
	private boolean runAgain 			= false;
	private boolean lockFormula 			= false;
	private boolean forceCalculation 	= false;

	List<String> parameterNamesList 		= null;
	private Icon gIcon			    		= null;
	private calibrationState pluginState = null; 

	private String calibrationFormula 	= "t-Rex";
	private String printFormula 		= "";
	private String selectedParameter 	= "";
	private String previousSampleName 	= "";
	private String hardSample 			= "";
	private String formulaCreatedDate 	= "";
	private String userName 				= "";

	public enum calibrationState {	  peaksOrManual , manualEntry, autoPeak, findingPeaks, addingMESF, finalized, propagateDrp, noParam		}

	@Override public String getName() {	return "Calibration";	}

	@Override public String getVersion() {	return "1.0";	}

	@Override public ExportFileTypes useExportFileType() { 	return null;		}

	@Override public List<String> getParameters() {	return null;		} 

	@Override public Icon getIcon()
	{
		if (calibrationState.findingPeaks == pluginState || calibrationState.addingMESF == pluginState)
		{
			URL url = getClass().getClassLoader().getResource("images/peaks.png");
			if (url != null)
				gIcon = new ImageIcon(url);
		}
		else if (calibrationState.manualEntry == pluginState)
		{
			URL url = getClass().getClassLoader().getResource("images/peaks.png");
			if (url != null)
				gIcon = new ImageIcon(url);
		}
		else if (calibrationState.finalized == pluginState || pluginState == calibrationState.propagateDrp)
		{
			URL url = getClass().getClassLoader().getResource("images/Regression.png");
			if (url != null)
				gIcon = new ImageIcon(url);
		}else if (calibrationState.noParam == pluginState)
		{
			URL url = getClass().getClassLoader().getResource("images/noParam.png");
			if (url != null)
				gIcon = new ImageIcon(url);
		}
		return gIcon;
	}


	@Override public SElement getElement() 
	{
		System.out.println("getElement");
		SElement result = new SElement(getName());
		
		result.setString("Parameters", selectedParameter);
		result.setString("calibrationFormula", calibrationFormula);
		result.setString("printFormula", printFormula);
		result.setString("previousSampleName", previousSampleName);
		result.setString("hardSample", hardSample);
		result.setString("userName", userName);
		result.setString("formulaCreatedDate", formulaCreatedDate);
		result.setBool("runAgain",runAgain);
		result.setBool("forceCalculation",forceCalculation);
		result.setBool("checkAuto",checkAuto);
		result.setBool("lockFormula",lockFormula);
		result.setInt("minimumGates", minimumGates);
		result.setInt("peaksFound", numPeaksFound);

		if(pluginState!=null) 
		{
			System.out.println("PluginState: "+pluginState.toString());
			result.setString("pluginState", pluginState.toString());
			System.out.println("Added: "+result.getString("pluginState"));
		}

		if(pluginState == calibrationState.manualEntry || pluginState == calibrationState.addingMESF && linRegAbs!=null && linRegMean!=null) 
		{
			SElement elementParentAbsolute = new SElement("AbsoluteValues");
			result.addContent(elementParentAbsolute);
			for (int i = 0; i < linRegAbs.length; i++) {
				SElement elementChildAbsolute = new SElement("absValuesElems");
				elementChildAbsolute.setDouble("Abs", linRegAbs[i]);	
				elementParentAbsolute.addContent(elementChildAbsolute);
			}

			SElement elementParentMean = new SElement("meanValues");
			result.addContent(elementParentMean);
			for (int i = 0; i < linRegMean.length; i++) 
			{
				SElement elementChildMeans = new SElement("meanVals");
				elementChildMeans.setDouble("means", linRegMean[i]);	
				elementParentMean.addContent(elementChildMeans);
			}

			SElement elementParameters = new SElement("parameterLister");
			result.addContent(elementParameters);
			for(String param: parameterNamesList)
			{
				SElement elementChildParams = new SElement("Pars");
				elementChildParams.setString("nam", param);	
				elementParameters.addContent(elementChildParams);
			}
		}
		return result;
	}


	@Override
	public void setElement(SElement element) 
	{
		calibrationFormula = element.getAttribute("calibrationFormula");
		printFormula = element.getAttribute("printFormula");
		pluginState 	= 		calibrationState.valueOf(element.getString("pluginState"));
		selectedParameter = 	element.getAttribute("Parameters");
		previousSampleName = element.getAttribute("previousSampleName");
		hardSample = 		element.getAttribute("hardSample");
		userName = 			element.getAttribute("userName");
		formulaCreatedDate = element.getAttribute("formulaCreatedDate");
		minimumGates = 		element.getInt("minimumGates", minimumGates);
		numPeaksFound = 		element.getInt("peaksFound", numPeaksFound);
		checkAuto = 			element.getBool("checkAuto");
		forceCalculation = 	element.getBool("forceCalculation");
		runAgain = 			element.getBool("runAgain");
		lockFormula = 		element.getBool("lockFormula");

		if( pluginState == calibrationState.manualEntry || pluginState == calibrationState.addingMESF && element.getChild("AbsoluteValues") != null) 
		{
			List<SElement> params = element.getChild("AbsoluteValues").getChildren();
			linRegAbs = new double[params.size()];
			for (int i = 0; i < params.size(); i++) {
				linRegAbs[i] = Double.parseDouble(params.get(i).getAttribute("Abs"));
			}

			List<SElement> params1 = element.getChild("meanValues").getChildren();
			linRegMean = new double[params.size()];		
			for (int i = 0; i < params1.size(); i++) {
				linRegMean[i] = Double.parseDouble(params1.get(i).getAttribute("means"));
			}

			List<SElement> params11 = element.getChild("parameterLister").getChildren();
			parameterNamesList = new ArrayList<String>();		
			for (int i = 0; i < params11.size(); i++) {
				parameterNamesList.add(params11.get(i).getAttribute("nam"));
				System.out.println(parameterNamesList.get(i));
			}
		}
	}

	public boolean applyToNewSamplePrompt(SElement fcmlQueryElement, List<String> parameterNames) {
		List<String> lst = FJPluginHelper.getParameterNames(fcmlQueryElement);
		List<String> params = new ArrayList<String>();
		for(String param: lst) 
		{
			params.add(ParameterUtil.stripStainName(param));
		}
		if(!params.contains(selectedParameter)) 
		{
			System.out.println("Samples: Parameter"+params);
			System.out.println("Target Parameter: "+selectedParameter);
			JOptionPane.showMessageDialog(null, selectedParameter+" is not in this sample's parameters.");
			pluginState = calibrationState.noParam;
			return false;
		}
		pluginState = calibrationState.propagateDrp;

		List<Object> guiObjects = new ArrayList<Object>();
		FJLabel explainTextFJLabel = new FJLabel(null, SwingConstants.CENTER);
		guiObjects.add(explainTextFJLabel);
		String textToDisplay = "";
		textToDisplay =  "<html><body>";
		textToDisplay += 	"<h2><center>				Calibrated Parameter: "+selectedParameter+"</center></h4>";
		textToDisplay += 	"<h5><center>				Created on "+formulaCreatedDate+"	</center></h4>";
		if(!userName.equalsIgnoreCase("DoNotApply"))
		{
			System.out.println("User name found: "+userName.equalsIgnoreCase("DoNotApply"));
//			textToDisplay += 	"<h5><center>				User: "+userName+"	</center></h4>";
		}
		textToDisplay += 	"<h5><center>				Formula:  "+printFormula+"	</center></h4>";
		textToDisplay += 	"<h5><center>				Sample: "+hardSample.replace(".csv", "").replace(".fcs", "").replace(".lmd", "").replace(".txt", "").trim()+"	</center></h4>";		

		if(lockFormula)
		{
			textToDisplay += 	"<ol type=\"i\">";
			textToDisplay += 		"<li><b>			Apply Formula.		  </b> 				   ";
			textToDisplay += 			"<ol type=\"i\">";
			textToDisplay += 				"<li>Calibrated Parameter will be applied to current sample 	   </li>  ";	
			textToDisplay += 				"<li>Node will remain locked 	   </li>  ";	
			textToDisplay += 			"</ol>";
			textToDisplay += 		"</li>";
			textToDisplay += 	"</ol>";
			textToDisplay += "</body></html>";
			explainTextFJLabel.setText(textToDisplay);

			Object[] optionsList = { "Do Not Apply", "Apply Formula" };
			int optionWindow = JOptionPane.showOptionDialog(
					null, 
					guiObjects.toArray(), 
					"Calibration Plugin",
					JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.PLAIN_MESSAGE, 
					null, optionsList, "Apply Formula" );
			if(optionWindow == Arrays.asList(optionsList).indexOf("Apply Formula")) 
			{
				System.out.println("User chose to apply formula, runAgain set to true");
				forceCalculation = true;
				runAgain = true;
				return true;
			}else 
				if(optionWindow == JOptionPane.CLOSED_OPTION || optionWindow== JOptionPane.CANCEL_OPTION || optionWindow == Arrays.asList(optionsList).indexOf("Do Not Apply"))
				{
//					JOptionPane.showMessageDialog(null, "Calibration was not applied");
					return false;
				}
		}
		else if(!lockFormula) {
			textToDisplay += 	"<ol type=\"i\">";
			textToDisplay += 		"<li>			Apply Formula.		    	 				   ";
			textToDisplay += 			"<ol type=\"i\">";
			textToDisplay += 				"<li>  Calibrated Parameter will be applied to current sample 	   </li>  ";	
			textToDisplay += 			"</ol>";
			textToDisplay += 		"</li>";
			textToDisplay += 		"<li>	 					Re-Calculate 								    ";
			textToDisplay += 			"<ol type=\"i\">";
			textToDisplay += 				"<li>			Re-Calculate Calibration 						</li>";	
			textToDisplay += 				"<li>			Choose to Auto-peak or Reference Points					</li>";	
			textToDisplay += 			"</ol>";
			textToDisplay += 		"</li>";
			textToDisplay +=			"<li><b>			Derived parameter can be applied to other samples. 	 		</b>	</li>";
			textToDisplay += 	"</ol>";
			textToDisplay += "</body></html>";
			explainTextFJLabel.setText(textToDisplay);
			Object[] optionsManualOrPeak = { "Apply Formula", "Re-Calculate", "Cancel" };
			int optionsJOptionPaneManualOrPeak = JOptionPane.showOptionDialog(
					null, 
					guiObjects.toArray(), 
					"Calibration Plugin",
					JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.PLAIN_MESSAGE, 
					null, optionsManualOrPeak, "Apply Formula");
			if(optionsJOptionPaneManualOrPeak == Arrays.asList(optionsManualOrPeak).indexOf("Apply Formula")) 
			{
				runAgain = true;
				return true;
			}else 
				if(optionsJOptionPaneManualOrPeak == Arrays.asList(optionsManualOrPeak).indexOf("Re-Calculate")) 
				{
					boolean retVal = initialPrompt(fcmlQueryElement, params);
					System.out.println("Initial prompt called, accepted? ->"+retVal);
					if(retVal) {		pluginState = null; }
					if(!retVal) {return false;}
				}else 
					if(optionsJOptionPaneManualOrPeak == Arrays.asList(optionsManualOrPeak).indexOf("Cancel") || optionsJOptionPaneManualOrPeak == JOptionPane.CLOSED_OPTION) 
					{
						return false;
					}
		}
		return false;
	}
	public boolean initialPrompt(SElement fcmlQueryElement, List<String> parameterNames) 
	{

		List<Object> guiObjects = new ArrayList<Object>();
		FJLabel explainTextFJLabel = new FJLabel(null, SwingConstants.CENTER);
		guiObjects.add(explainTextFJLabel);
		String textToDisplay = "";
		textToDisplay =  "<html><body>";
		textToDisplay =  "<html><body>";
		textToDisplay += 	"<h2><center>				Calibration Steps					</center></h2>";
		textToDisplay += 	"<ol>";
		textToDisplay += 		"<li><b>			Select parameter to be calibrated			    	 	   	</b>	</li>  ";
		textToDisplay += 		"<li><b>			Choose Manual Entry or Auto-Peak finder.		    	 		</b>		   ";
		textToDisplay += 			"<ol>";
		textToDisplay += 				"<li>			Auto-Peak: Automagically find peaks in parameter. 	</li>  ";	
		textToDisplay += 				"<li>			Reference Points: Manually enter reference points.	</li>  ";	
		textToDisplay += 			"</ol>";
		textToDisplay += 		"</li>";
		textToDisplay += 		"<li>	 		Enter data points. 								    ";
		textToDisplay += 			"<ol>";
		textToDisplay += 				"<li>			Linear regression will be perfomed. 							</li>";	
		textToDisplay += 				"<li>			Derived parameter will be calculated.					</li>";	
		textToDisplay += 			"</ol>";
		textToDisplay += 		"</li>";
		textToDisplay +=			"<li>			Derived parameter can be applied to other samples. 	 			</li>";
		textToDisplay += 	"</ol>";
		textToDisplay += "</body></html>";
		textToDisplay += "</body></html>";
		explainTextFJLabel.setText(textToDisplay);
		// Combo Drop-down of Parameters
		ArrayList<String> parameters = new ArrayList<String>();
		for (Object s : parameterNames.toArray()) {parameters.add((String) s);}
		FJComboBox parameterCB1 = new FJComboBox(parameters.toArray());
		FJPanel parameter1Panel = new FJPanel();
		parameterCB1.setName("ParameterComboBox");
		HBox hboxParameterCB1 = new HBox(Box.createHorizontalGlue(), parameterCB1, Box.createHorizontalGlue());
		parameter1Panel.add(hboxParameterCB1);
		HBox hboxPanels = new HBox(Box.createHorizontalGlue(),parameter1Panel, Box.createHorizontalGlue());
		guiObjects.add(hboxPanels);
		Object[] options = { "Auto-Peak", "Reference Points", "Cancel" };
		int jOptionWindow = JOptionPane.showOptionDialog(
				null, 
				guiObjects.toArray(), 
				"Calibration Plugin",
				JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.PLAIN_MESSAGE, 
				null, options, "Auto-Peak");

		selectedParameter = ParameterUtil.stripStainName(parameterCB1.getSelectedItem().toString());
		if(jOptionWindow == Arrays.asList(options).indexOf("Auto-Peak")) 
		{
			boolean retVal = promptAutoPeaks(fcmlQueryElement, parameterNames);
			return retVal;
		}
		else 
			if(jOptionWindow == Arrays.asList(options).indexOf("Reference Points")) 
			{
				boolean retVal = peakNumSelector(fcmlQueryElement, parameterNames);
				return retVal;
			}
			else 
				if(jOptionWindow == Arrays.asList(options).indexOf("Cancel") || jOptionWindow == JOptionPane.CLOSED_OPTION) 
				{
					return false;
				}
		return false;
	}
	public boolean promptAutoPeaks(SElement fcmlQueryElement, List<String> parameterNames) 
	{
		{
			{
				List<Object> guiObjects = new ArrayList<Object>();
				// Auto-peak option
				FJLabel explainTextFJLabel = new FJLabel(null, SwingConstants.CENTER);
				explainTextFJLabel = new FJLabel();
				String textToDisplay = "";
				guiObjects.add(explainTextFJLabel);
				textToDisplay = "";
				textToDisplay =  "<html><body>";
				textToDisplay += 	"<h2><center>				Calibration Steps					</center></h2>";
				textToDisplay += 	"<ol>";
				textToDisplay += 		"<li>			Select parameter to be calibrated			    	 	   		</li>  ";
				textToDisplay += 		"<li>			Choose Manual Entry or Auto-Peak finder.		    	 				   ";
				textToDisplay += 			"<ol>";
				textToDisplay += 				"<li>			Auto-Peak: Automagically find peaks in parameter. 	</li>  ";	
				textToDisplay += 				"<li>			Reference Points: Manually enter reference points.	</li>  ";	
				textToDisplay += 			"</ol>";
				textToDisplay += 		"</li>";
				textToDisplay += 		"<li><b>	 		Enter data points. 		</b>						    ";
				textToDisplay += 			"<ol>";
				textToDisplay += 				"<li>			Linear regression performed. 							</li>";	
				textToDisplay += 				"<li>			Derived parameter will be calculated.					</li>";	
				textToDisplay += 			"</ol>";
				textToDisplay += 		"</li>";
				textToDisplay +=			"<li>			Derived parameter can be applied to other samples. 	 			</li>";
				textToDisplay += 	"</ol>";
				textToDisplay += "</body></html>";
				explainTextFJLabel.setText(textToDisplay);

				PeakList pks = peakfinder(fcmlQueryElement, selectedParameter);
				System.out.println("Amount of Peaks Finder: "+pks.size());
				int num = pks.size();
				RangedDoubleTextField[] peakFieldArray = new RangedDoubleTextField[num];
				HBox peakHBox = null;

				linRegMean = new double[num];
				linRegAbs = new double[num];

				int i = 0;
				if(num<2) {
					JOptionPane.showMessageDialog(null, "Found "+num+" Peaks, retry with manual or different parameter.");
					boolean retVal = initialPrompt(fcmlQueryElement, parameterNames);
					return retVal;
				}
				for(PeakInfo peakInf: pks) 
				{
					double mean = FJPluginHelper.channelToScale(PluginHelper.getSampleURI(fcmlQueryElement), selectedParameter,peakInf.getMean(), histoSize);
					DecimalFormat f = new DecimalFormat("##.00");
					String mean2Dec = f.format(mean);
					FJLabel minPeakLabel = new FJLabel("Peak_"+i+" Mean: "+mean2Dec);
					String minPeakTip = "Add absolute value for Peak_"+i+ ". Must be less than 100M";
					minPeakLabel.setToolTipText(minPeakTip);
					peakFieldArray[i] = new RangedDoubleTextField(-100000000, 100000000, new IntlFormatter(2));
					peakFieldArray[i].setDouble(0.00);
					peakFieldArray[i].setToolTipText(minPeakTip);
					GuiFactory.setSizes(peakFieldArray[i], new Dimension(70, 25));
					peakHBox = new HBox(Box.createHorizontalGlue(), minPeakLabel, peakFieldArray[i], Box.createHorizontalGlue());
					guiObjects.add(peakHBox);
					linRegMean[i] =  Double.parseDouble(mean2Dec);
					i++;
				}
				JCheckBox CreateGateCheck = new JCheckBox("Create-Gates in Workspace");
				CreateGateCheck.setSelected(true);
				CreateGateCheck.setToolTipText("Checking true will create child populations for each gate found.");
				HBox createGatesBox = new HBox(Box.createHorizontalGlue(), CreateGateCheck, Box.createHorizontalGlue());
				guiObjects.add(createGatesBox);

				JCheckBox lockFormulaCheck = new JCheckBox("Lock Calibration Formula");
				lockFormulaCheck.setToolTipText("Calibration formula cannot be altered after creation. Only making a new node will override it.");
				lockFormulaCheck.setSelected(false);
				HBox lockFormulaCheckBox = new HBox(Box.createHorizontalGlue(), lockFormulaCheck, Box.createHorizontalGlue());
				guiObjects.add(lockFormulaCheckBox);

				Object[] options = {  "Calibrate","Go Back", "Cancel" }; 
				int jOptionWindow = JOptionPane.showOptionDialog(
						null, 
						guiObjects.toArray(), 
						"Calibration Plugin",
						JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.PLAIN_MESSAGE, 
						null, options, "Calibrate");
				if (jOptionWindow == Arrays.asList(options).indexOf("Calibrate")) 
				{
					pluginState = calibrationState.addingMESF;
					checkAuto = CreateGateCheck.isSelected();
					lockFormula = lockFormulaCheck.isSelected();

					for (int i1 = 0; i1 < peakFieldArray.length; i1++) 
					{
						linRegAbs[i1] =  peakFieldArray[i1].getDouble();
						System.out.println("Peak_"+i1+" absolute values collected: "+linRegAbs[i1]+" to Mean: "+linRegMean[i1]);
					}
					hardSample = FJPluginHelper.getSampleName(fcmlQueryElement);
					runAgain = true;
					if(lockFormula) 
					{
//						boolean retVal = askUserName( fcmlQueryElement, parameterNames, false);
						return true;
					}
					return true;
				}
				else 
					if(jOptionWindow == Arrays.asList(options).indexOf("Go Back"))
					{
						boolean retVal = initialPrompt(fcmlQueryElement, parameterNames);
						return retVal;
					}
					else 
						if(jOptionWindow == Arrays.asList(options).indexOf("Cancel") || jOptionWindow == JOptionPane.CLOSED_OPTION) 
						{
							return false;
						}
			}
		}
		return false;
	}
	public boolean peakNumSelector(SElement fcmlQueryElement, List<String> parameterNames) 
	{
		List<Object> guiObjects = new ArrayList<Object>();
		// Auto-peak option
		FJLabel explainTextFJLabel = new FJLabel(null, SwingConstants.CENTER);
		String textToDisplay = "";
		pluginState = calibrationState.peaksOrManual;
		explainTextFJLabel = new FJLabel();
		guiObjects.add(explainTextFJLabel);
		textToDisplay = "";
		textToDisplay =  "<html><body>";
		textToDisplay += 	"<h2><center>				Calibration Steps					</center></h2>";
		textToDisplay += 	"<ol>";
		textToDisplay += 		"<li>			Select parameter to be calibrated			    	 	   		</li>  ";
		textToDisplay += 		"<li>			Choose Manual Entry or Auto-Peak finder.		    	 				   ";
		textToDisplay += 			"<ol>";
		textToDisplay += 				"<li>			Auto-Peak: Automagically find peaks in parameter. 	</li>  ";	
		textToDisplay += 				"<li>			Reference Points: Manually enter reference points.	</li>  ";	
		textToDisplay += 			"</ol>";
		textToDisplay += 		"</li>";
		textToDisplay += 		"<li><b>	 		Enter data points. 		</b>						    ";
		textToDisplay += 			"<ol>";
		textToDisplay += 				"<li>			Linear regression performed. 							</li>";	
		textToDisplay += 				"<li>			Derived parameter will be calculated.					</li>";	
		textToDisplay += 			"</ol>";
		textToDisplay += 		"</li>";
		textToDisplay +=			"<li>			Derived parameter can be applied to other samples. 	 			</li>";
		textToDisplay += 	"</ol>";
		textToDisplay += "</body></html>";
		explainTextFJLabel.setText(textToDisplay);
		// Manual Entry
		FJLabel minPeakLabel = new FJLabel("# of Fields: ");
		String minPeakTip = "Peaks in sample to detect";
		minPeakLabel.setToolTipText(minPeakTip);
		RangedIntegerTextField minPeakField = new RangedIntegerTextField(0, 50);
		minPeakField.setInt(8);
		minPeakField.setToolTipText(minPeakTip);
		GuiFactory.setSizes(minPeakField, new Dimension(50, 25));
		HBox peakHBox3 = new HBox(Box.createHorizontalGlue(), minPeakLabel, minPeakField, Box.createHorizontalGlue());
		guiObjects.add(peakHBox3);
		Object[] options = { "Set Number of Rows", "Go Back",  "Cancel"};
		int jOptionWindow = JOptionPane.showOptionDialog(null, 
				guiObjects.toArray(), 
				"Calibration Plugin",
				JOptionPane.YES_NO_CANCEL_OPTION, 
				JOptionPane.PLAIN_MESSAGE, 
				null, options, "Set Number of Rows");
		if(jOptionWindow == Arrays.asList(options).indexOf("Set Number of Rows")) 
		{
			boolean retVal = promptManualPeakFields(fcmlQueryElement, parameterNames,minPeakField.getInt());
			if(retVal) {return true;}
		}else 
			if(jOptionWindow == Arrays.asList(options).indexOf("Go Back")) 
			{
				boolean retVal = initialPrompt(fcmlQueryElement, parameterNames);
				if(!retVal) {return false;}
			}
			else 
				if(jOptionWindow == Arrays.asList(options).indexOf("Cancel") || jOptionWindow == JOptionPane.CLOSED_OPTION) 
				{
					return false;
				}
		return false;
	}
	public boolean askUserName(SElement fcmlQueryElement, List<String> parameterNames, boolean manual)
	{
		if(pluginState==null) 
		{
			pluginState = calibrationState.findingPeaks;
		}
		List<Object> guiObjects = new ArrayList<Object>();
		FJLabel explainTextFJLabel = new FJLabel(null, SwingConstants.CENTER);
		guiObjects.add(explainTextFJLabel);
		String textToDisplay = "";
		textToDisplay =  "<html><body>";
		textToDisplay =  "<html><body>";
		textToDisplay += 	"<h2><center>				Calibration Lock					</center></h2>";
		textToDisplay += 	"<ol>";
		textToDisplay += 		"<li><b>			Add User Name			    	 	   	</b>	</li>  ";
		textToDisplay += 		"<li><b>			Date and Sample name will be hardcoded to calibration node.		    	 		</b>		   ";
		textToDisplay += 		"<li><b>			Calibration node can be applied to other samples.		    	 		</b>		   ";
		textToDisplay += 	"</ol>";
		textToDisplay += "</body></html>";
		textToDisplay += "</body></html>";
		explainTextFJLabel.setText(textToDisplay);
		// Combo Drop-down of Parameters

		String tooltip = "User name to save to locked Calibration node";
		FJLabel runIDLabel = new FJLabel("Username: ");
		FJTextField bsRunIDField = new FJTextField();
		bsRunIDField.setText("Add User Name");
		bsRunIDField.setToolTipText(tooltip);
		GuiFactory.setSizes(bsRunIDField, new Dimension(200, 25));
		HBox runIDBox = new HBox(Box.createHorizontalGlue(), runIDLabel, bsRunIDField, Box.createHorizontalGlue());
		guiObjects.add(runIDBox);

		Object[] options = { "Save","Go Back", "Cancel" };
		int jOptionWindow = JOptionPane.showOptionDialog(
				null, 
				guiObjects.toArray(), 
				"Calibration Plugin",
				JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.PLAIN_MESSAGE, 
				null, options, "Save");

		if(jOptionWindow == Arrays.asList(options).indexOf("Save")) 
		{
			userName = bsRunIDField.getText();
			if(userName.equalsIgnoreCase("Add User Name")) 
			{
				JOptionPane.showMessageDialog(null, "No user name entered, a username will not be recorded");
				userName="DoNotApply";
			}
			return true;
		}else 
			if(jOptionWindow == Arrays.asList(options).indexOf("Go Back")) 
			{
				if(manual) {
					boolean retBool = peakNumSelector(fcmlQueryElement, parameterNames);
					return retBool;
				}else if(!manual) {
					boolean retVal = promptAutoPeaks(fcmlQueryElement, parameterNames);
					return retVal;
				}
			}
			else 
				if(jOptionWindow == Arrays.asList(options).indexOf("Cancel") || jOptionWindow == JOptionPane.CLOSED_OPTION) 
				{
					return false;
				}
		return false;
	}
	public boolean promptManualPeakFields(SElement fcmlQueryElement, List<String> parameterNames, int minPeak) 
	{			 
		List<Object> guiObjects = new ArrayList<Object>();
		// Auto-peak option
		pluginState = calibrationState.peaksOrManual;
		FJLabel explainTextFJLabel = new FJLabel(null, SwingConstants.CENTER);
		explainTextFJLabel = new FJLabel();
		String textToDisplay = "";
		guiObjects.add(explainTextFJLabel);
		textToDisplay = "";
		textToDisplay =  "<html><body>";
		textToDisplay += 	"<h2><center>				Calibration Steps					</center></h2>";
		textToDisplay += 	"<ol>";
		textToDisplay += 		"<li>			Select parameter to be calibrated			    	 	   		</li>  ";
		textToDisplay += 		"<li>			Choose Manual Entry or Auto-Peak finder.		    	 				   ";
		textToDisplay += 			"<ol>";
		textToDisplay += 				"<li>			Auto-Peak: Automagically find peaks in parameter. 	</li>  ";	
		textToDisplay += 				"<li>			Reference Points: Manually enter reference points.	</li>  ";	
		textToDisplay += 			"</ol>";
		textToDisplay += 		"</li>";
		textToDisplay += 		"<li><b>	 		Enter data points. 		</b>						    ";
		textToDisplay += 			"<ol>";
		textToDisplay += 				"<li>			Linear regression performed. 							</li>";	
		textToDisplay += 				"<li>			Derived parameter will be calculated.					</li>";	
		textToDisplay += 			"</ol>";
		textToDisplay += 		"</li>";
		textToDisplay +=			"<li>			Derived parameter can be applied to other samples. 	 			</li>";
		textToDisplay += 	"</ol>";
		textToDisplay += "</body></html>";
		explainTextFJLabel.setText(textToDisplay);
		int num = minPeak;
		HBox peakHBox = null;
		HBox peakHBox2 = null;
		RangedDoubleTextField[] peakFieldArray = new RangedDoubleTextField[num];
		RangedDoubleTextField[] peakFieldArray2 = new RangedDoubleTextField[num];
		linRegMean = new double[num];
		linRegAbs = new double[num];
		FJLabel minPeakLabel1 = new FJLabel("Peak # ");
		FJLabel minPeakLabel11 = new FJLabel("Mean Values");
		FJLabel minPeakLabel2 = new FJLabel("Absolute Values");

		peakHBox2 = new HBox(Box.createHorizontalGlue(), minPeakLabel1,minPeakLabel11, minPeakLabel2, Box.createHorizontalGlue());
		guiObjects.add(peakHBox2);

		for (int i = 0; i < minPeak; i++) {

			FJLabel minPeakLabel0 = new FJLabel("Peak "+i);
			String minPeakTip2 = "Add mean value for Peak "+i+ ". Must be less than 100M";

			peakFieldArray2[i] = new RangedDoubleTextField(-100000000, 100000000, new IntlFormatter(2));
			peakFieldArray2[i].setDouble(0.00);
			peakFieldArray2[i].setToolTipText(minPeakTip2);
			GuiFactory.setSizes(peakFieldArray2[i], new Dimension(70, 25));
			String minPeakTip1 = "Add absolute value for Peak_"+i+ ". Must be less than 100M";

			minPeakLabel11.setToolTipText(minPeakTip1);
			peakFieldArray[i] = new RangedDoubleTextField(-100000000, 100000000, new IntlFormatter(2));
			peakFieldArray[i].setDouble(0.00);
			peakFieldArray[i].setToolTipText(minPeakTip1);
			GuiFactory.setSizes(peakFieldArray[i], new Dimension(70, 25));
			peakHBox = new HBox(Box.createHorizontalGlue(), minPeakLabel0,peakFieldArray2[i], peakFieldArray[i], Box.createHorizontalGlue());
			guiObjects.add(peakHBox);
		}
		JCheckBox lockFormulaCheck = new JCheckBox("Lock Calibration Formula");
		lockFormulaCheck.setToolTipText("Calibration formula cannot be altered after creation. Only making a new node will override it.");
		lockFormulaCheck.setSelected(false);
		HBox lockFormulaCheckBox = new HBox(Box.createHorizontalGlue(), lockFormulaCheck, Box.createHorizontalGlue());
		guiObjects.add(lockFormulaCheckBox);

		Object[] options = { "Calibrate", "Go Back",  "Cancel"};
		int jOptionWindow = JOptionPane.showOptionDialog(null, 
				guiObjects.toArray(), 
				"Calibration Plugin",
				JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.PLAIN_MESSAGE, 
				null, options, options[0]);
		if(jOptionWindow == Arrays.asList(options).indexOf("Calibrate")) 
		{
			pluginState = calibrationState.manualEntry;
			for (int i = 0; i <  num; i++) {
				linRegMean[i] = peakFieldArray2[i].getDouble();
				linRegAbs[i] = peakFieldArray[i].getDouble();
				System.out.println(linRegMean[i]);
				System.out.println(linRegAbs[i]);
			}
			hardSample = FJPluginHelper.getSampleName(fcmlQueryElement);
			lockFormula = lockFormulaCheck.isSelected();
			runAgain = true;
			if(lockFormula) 
			{
//				return askUserName( fcmlQueryElement,  parameterNames, true);
				return true;
			}
			return true;
		}else 
			if(jOptionWindow == Arrays.asList(options).indexOf("Go Back")) 
			{
				boolean retVal = peakNumSelector(fcmlQueryElement, parameterNames);
				if(!retVal) {return false;}
			}
			else 
				if(jOptionWindow == Arrays.asList(options).indexOf("Cancel") || jOptionWindow == JOptionPane.CLOSED_OPTION) 
				{
					return false;
				}
		return false;
	}
	@Override
	public boolean promptForOptions(SElement fcmlQueryElement, List<String> parameterNames) 
	{
		if(pluginState==null) 
		{
			parameterNamesList = new ArrayList<String>();
			for (int i = 0; i < parameterNames.size(); i++) {
				String param = ParameterUtil.stripStainName(parameterNames.get(i).toString());
				parameterNamesList.add(param);
			}
			boolean retvals = initialPrompt( fcmlQueryElement,  parameterNames);

			return retvals;
		}
		else 
			if(pluginState == calibrationState.propagateDrp && calibrationFormula != "t-Rex") 
			{
				System.out.println("Entered promptForOptions in propagate mode");
				boolean retVal = applyToNewSamplePrompt(fcmlQueryElement,  parameterNames);
				if(retVal) {
					System.out.println("Apply to sample: "+retVal);
					forceCalculation=true;}
				return retVal;
			}else if(pluginState == calibrationState.noParam) {
				JOptionPane.showMessageDialog(null, "Parameter to calibrate is not available or written verbatim in current sample.");
				return false;
			}
		return false;
	}

	/*
	 * This method uses the fcmlQueryElement passed in by FlowJo to find peaks 
	 * in the selected parameter. 
	 * @param fcmlQueryElement of parent node of plugin.
	 * @param mParameter, Parameter to find peaks in.
	 */
	private PeakList peakfinder(SElement fcmlQueryElement, String mParameter) 
	{
		// Get histogram of calibrated parameter using 4096 channels.
		double[] histogramData = FJPluginHelper.getHistogram(fcmlQueryElement, selectedParameter, true, histoSize);
		// clear edges of histograms
		histogramData[0] = 0;
		histogramData[1] = 0;
		histogramData[histoSize - 1] = 0;
		histogramData[histoSize - 2] = 0;
		// Create a discrete function from the histogram array.
		DiscreteFunction DF = new DiscreteFunction(histogramData);
		// Create a PeakFinder object using the discrete function.
		PeakFinder PF = new PeakFinder(DF);
		// Create a PeakList to store the peaks found from PeakFinder().
		PeakList peakList = PF.findPeaks();
		return peakList;
	}
	/*
	 *  This method creates FlowJo's XML gating markup language (Gating-ML)
	 *  It is passed a Peak list, iterates over each peak and creates a new 
	 *  peak over the parameter to calibrate.
	 *  @param ExternalAlgorithmResults
	 *  @param int clusters
	 *  @param String runID
	 */
	private void addGatingML(ExternalAlgorithmResults result, SElement fcmlQueryElement, PeakList peakList, String parameterToCalibrate, String formula)
	{
		SElement gate = new SElement("gating:Gating-ML");
		int count = 0;
		for(PeakInfo pk: peakList) 
		{
			// Find mix, max, and mean from current peaks.
			double minX = FJPluginHelper.channelToScale(PluginHelper.getSampleURI(fcmlQueryElement), selectedParameter,pk.getMin(), histoSize);
			double maxX = FJPluginHelper.channelToScale(PluginHelper.getSampleURI(fcmlQueryElement), selectedParameter,pk.getMax(), histoSize);
			//			double mean = FJPluginHelper.channelToScale(PluginHelper.getSampleURI(fcmlQueryElement), selectedParameter,pk.getMean(), histoSize);

			// create the XML elements for a 1-D range gate
			SElement rangeGateElem = new SElement("gating:RectangleGate");
			rangeGateElem.setString("isTinted", "1");
			rangeGateElem.setString("gating:id", "Peak_"+count);
			gate.addContent(rangeGateElem);

			// create the dimension XML element
			SElement dimElem = new SElement("gating:dimension");
			dimElem.setDouble("gating:min", minX);
			dimElem.setDouble("gating:max", maxX);
			rangeGateElem.addContent(dimElem);

			// create the parameter name XML element
			SElement fcsDimElem = new SElement("data-type:fcs-dimension");
			fcsDimElem.setString("data-type:name", selectedParameter);
			dimElem.addContent(fcsDimElem);			
			count++;
		}
		// Add gatingML as a query to the ExternalAlgorithmResults object. 
		// Does not need to be returned. 
		result.setGatingML(gate.toString());
	}

	@Override
	public ExternalAlgorithmResults invokeAlgorithm(SElement fcmlQueryElement, File arg1, File arg2) 
	{

		ExternalAlgorithmResults xmlResult = new ExternalAlgorithmResults();
		
		String currentSample = FJPluginHelper.getSampleName(fcmlQueryElement);

//		System.out.println("Entered Invoke Algorithm: "+pluginState);
//		System.out.println("Run Again? -> "+runAgain);
//		System.out.println("Same sample: "+currentSample.equalsIgnoreCase(previousSampleName));
//		System.out.println("Current Sample: "+currentSample);
//		System.out.println("Previous Sample: "+previousSampleName);
//		System.out.println("Original Sample: "+hardSample);
//		System.out.println("Force Recalculate?: "+forceCalculation);
//		

		if(pluginState==null) 
		{
			pluginState = calibrationState.findingPeaks;
		}
		
		if(runAgain) 
		{
			if (pluginState == calibrationState.manualEntry)
			{
				System.out.println("Entered Manual Entry");
				if(linRegAbs==null || linRegMean==null) 
				{
					System.out.println("linRegAbs null");
					return xmlResult;
				}
				LinearRegression simpleRegression = new LinearRegression(linRegMean, linRegAbs);
				for (int i = 0; i < linRegAbs.length; i++) {
					System.out.println(linRegMean[i]+ " "+linRegAbs[i]);

				}
				String formula = "t-rex";
				String printingFormula = "";
				double simpIntercept = 0;
				double simpSlope = 0;
				String formattedIntercept = "";
				String formattedSlope = "";
				if(simpleRegression.intercept()>=0) 
				{
					System.out.println("Entered linearRegression analysis");
					simpIntercept =simpleRegression.intercept();
					formattedIntercept = String.format ("%.10f", simpIntercept);
					simpSlope =simpleRegression.slope();
					if(simpSlope<0) {
						simpSlope = simpSlope*-1;
					}
					formattedSlope = String.format ("%.10f", simpSlope);
					formula = selectedParameter+"*"+formattedSlope+"+"+formattedIntercept;
					System.out.println("Calibration Formula: "+formula);
					formattedSlope = String.format ("%.4f", simpSlope);
					formattedIntercept = String.format ("%.4f", simpIntercept);
					printingFormula = selectedParameter+"*"+formattedSlope+"+"+formattedIntercept;
				}
				else
					if(simpleRegression.intercept()<0) {
						simpIntercept =(simpleRegression.intercept()*-1);
						formattedIntercept = String.format ("%.10f", simpIntercept);
						simpSlope =(simpleRegression.slope()*-1);
						if(simpSlope<0) {
							simpSlope = simpSlope*-1;
						}
						formattedSlope = String.format ("%.10f", simpSlope);
						formula = selectedParameter+"*"+formattedSlope+"-"+formattedIntercept;
						formattedSlope = String.format ("%.4f", simpSlope);
						formattedIntercept = String.format ("%.4f", simpIntercept);
						printingFormula = selectedParameter+"*"+formattedSlope+"-"+formattedIntercept;
					}
				if(formula=="t-rex") 
				{
					JOptionPane.showMessageDialog(null, "Linear regression not applied");
				}
				else
					if(formula!="t-rex")
					{
						if(simpIntercept!=0 && simpSlope!=0) 
						{
							calibrationFormula = formula;

							if(lockFormula) 
							{
								System.out.println("Lock formula selected in manual entry");
								xmlResult.addDerivedParameterFormula("Calibrated_"+selectedParameter, calibrationFormula);
								JOptionPane.showMessageDialog(null, "Lock Formula selected, drag-n-drop plugin node to calibrate other samples. R^2 "+simpleRegression.R2());
								DateTimeFormatter dtf = DateTimeFormatter.ofPattern("yyyy/MM/dd HH:mm:ss");
								LocalDateTime now = LocalDateTime.now();
								System.out.println(dtf.format(now)); //2016/11/16 12:08:43
								formulaCreatedDate = dtf.format(now);
								runAgain = false;
							}
							else 
								if (!lockFormula) 
								{
									System.out.println("Non-locked entered in manual entry mode.");
									Sample sample = FJPluginHelper.getSample(fcmlQueryElement);

									SElement elem = new SElement(FEML.DerivedParameter);
									elem.setString(FEML.name, "Calibrated_"+selectedParameter);
									elem.setString(FJML.type, FEML.ALGEBRAIC_NAME);
//									elem.setString(arg0, arg1);
									String formattedFormula = "";
									if(simpleRegression.intercept()<0) 
									{
										formattedFormula= "<Param name=\""+selectedParameter+"\" />*"+formattedSlope+"-"+formattedIntercept;
									}
									if(simpleRegression.intercept()>=0) 
									{
										formattedFormula= "<Param name=\""+selectedParameter+"\" />*"+formattedSlope+"+"+formattedIntercept;
									}
									calibrationFormula = formattedFormula;
									elem.setString(FJML.formula, formattedFormula);
									DerivedParameter param = new AlgebraicParameter(FJML.EventNumberDP, elem);
									param.setAttributes(elem, true);
									System.out.println("Transform: "+sample.getParameter(selectedParameter).getTransform().toString());
									elem.addContent(param.toString());

									System.out.println(elem.toString());
									DerivedParameterNode node = new DerivedParameterNode(sample.getWorkspace(), elem);
									node.getParameter().setTransform(sample.getParameter(selectedParameter).getTransform());
									sample.getSampleNode().addOrReplaceDerivedParameterNode(node);

									DateTimeFormatter dtf = DateTimeFormatter.ofPattern("yyyy/MM/dd HH:mm:ss");
									LocalDateTime now = LocalDateTime.now();
									formulaCreatedDate = dtf.format(now);
									JOptionPane.showMessageDialog(null, "Linear regression applied. Formula "+printingFormula+". R^2"+simpleRegression.R2());
								}
							printFormula = printingFormula;
							xmlResult.setWorkspaceString(printingFormula);
							pluginState = calibrationState.propagateDrp;
							runAgain = false;
							return xmlResult;
						}
						else 
							if(simpIntercept==0 && simpSlope==0) 
							{
								checkAuto = false;
								JOptionPane.showMessageDialog(null, "Intercept and Slope were zeros, retry or change parameter.");
								return xmlResult;
							}
					}
			}
			if (pluginState == calibrationState.addingMESF)
			{
				xmlResult.setWorkspaceString("FindingPeaks");
				System.out.println("Checkbox " +checkAuto);
				PeakList pks = null;
				if(checkAuto) 
				{
					pks = peakfinder(fcmlQueryElement, selectedParameter);
					numPeaksFound = pks.size();
				}
				if(linRegAbs==null || linRegMean==null || linRegAbs.length<1)
				{
					JOptionPane.showMessageDialog(null, "Internally could not collect absolute values and means");
					return xmlResult;
				}
				LinearRegression regressionAnalysis = new LinearRegression(linRegMean, linRegAbs);
				String formula = "t-rex";
				String printingFormula = "";
				String formattedIntercept ="";
				String formattedSlope = "";
				double simpIntercept = 0;
				double simpSlope = 0;
				if(regressionAnalysis.intercept()>=0) {
					simpIntercept =regressionAnalysis.intercept();
					formattedIntercept = String.format ("%.10f", simpIntercept);
					simpSlope =regressionAnalysis.slope();
					if(simpSlope <0) {
						simpSlope = simpSlope*-1;
					}
					formattedSlope = String.format ("%.10f", simpSlope);
					formula = selectedParameter+"*"+formattedSlope+"+"+formattedIntercept;
					formattedSlope = String.format ("%.4f", simpSlope);
					formattedIntercept = String.format ("%.4f", simpIntercept);
					printingFormula = selectedParameter+"*"+formattedSlope+"+"+formattedIntercept;
				}
				else 
					if(regressionAnalysis.intercept()<0) 
					{
						simpIntercept =(regressionAnalysis.intercept()*-1);
						formattedIntercept = String.format ("%.10f", simpIntercept);
						if(regressionAnalysis.slope() <0) 
						{
							simpSlope = simpSlope*-1;
						}
						simpSlope = regressionAnalysis.slope();
						formattedSlope = String.format ("%.10f", simpSlope);
						formula = selectedParameter+"*"+formattedSlope+"-"+formattedIntercept;
						formattedSlope = String.format ("%.4f", simpSlope);
						formattedIntercept = String.format ("%.4f", simpIntercept);
						printingFormula = selectedParameter+"*"+formattedSlope+"-"+formattedIntercept;
					}
				if(formula=="t-rex") 
				{
					JOptionPane.showMessageDialog(null, "Could not compute linear regression.");
					return xmlResult;
				}
				else 		
					if(simpIntercept==0 && simpSlope==0) 
					{
						JOptionPane.showMessageDialog(null, "Intercept and Slope were zeros, retry or change parameter.");
						boolean retVal = initialPrompt(fcmlQueryElement, parameterNamesList);
						if(retVal) {return xmlResult;}
						else if(!retVal) { return xmlResult; }
					}
					else if(simpIntercept!=0 && simpSlope!=0) 
					{

						if(checkAuto || pks!=null) 
						{
							addGatingML(xmlResult, fcmlQueryElement, pks, selectedParameter, calibrationFormula);
						}
						calibrationFormula = formula;
						System.out.println("Lock Formula: "+lockFormula);
						if(lockFormula) 
						{
							System.out.println("Lock Formula true, added derived parameter to plugin node.");
							JOptionPane.showMessageDialog(null, "Lock Formula selected, drag-n-drop plugin node to calibrate other samples. R^2 "+regressionAnalysis.R2());
							xmlResult.addDerivedParameterFormula("Calibrated_"+selectedParameter, calibrationFormula);

							DateTimeFormatter dtf = DateTimeFormatter.ofPattern("yyyy/MM/dd HH:mm:ss");
							LocalDateTime now = LocalDateTime.now();
							System.out.println(dtf.format(now)); //2016/11/16 12:08:43
							formulaCreatedDate = dtf.format(now);
							runAgain = false;
						}else 
							if (!lockFormula)  
							{
								SElement elem = new SElement(FEML.DerivedParameter);
								elem.setString(FEML.name, "Calibrated_"+selectedParameter);
								elem.setString(FJML.type, FEML.ALGEBRAIC_NAME);								
								String formattedFormula = "";
								if(regressionAnalysis.intercept()<0) 
								{
									formattedFormula= "<Param name=\""+selectedParameter+"\" />*"+formattedSlope+"-"+formattedIntercept;
								}
								if(regressionAnalysis.intercept()>=0) 
								{
									formattedFormula= "<Param name=\""+selectedParameter+"\" />*"+formattedSlope+"+"+formattedIntercept;
								}
								calibrationFormula = formattedFormula;
								elem.setString(FJML.formula, formattedFormula);
								DerivedParameter param = new AlgebraicParameter(FJML.EventNumberDP, elem);
								param.setAttributes(elem, true);
								elem.addContent(param.toString());

								System.out.println(elem.toString());
								Sample sample = FJPluginHelper.getSample(fcmlQueryElement);
								DerivedParameterNode node = new DerivedParameterNode(sample.getWorkspace(), elem);
								node.getParameter().setTransform(sample.getParameter(selectedParameter).getTransform());
								sample.getSampleNode().addOrReplaceDerivedParameterNode(node);

								DateTimeFormatter dtf = DateTimeFormatter.ofPattern("yyyy/MM/dd HH:mm:ss");
								LocalDateTime now = LocalDateTime.now();
								System.out.println(dtf.format(now));
								formulaCreatedDate = dtf.format(now);
								JOptionPane.showMessageDialog(null, "Derived parameter created: "+printingFormula+". R^2 "+regressionAnalysis.R2());
							}
						xmlResult.setWorkspaceString(printingFormula);
						pluginState = calibrationState.propagateDrp;
						runAgain = false;
						printFormula = printingFormula;
						previousSampleName = FJPluginHelper.getSampleName(fcmlQueryElement);
						System.out.println("Return result: "+pluginState);
						return xmlResult;
					}
			}else if(forceCalculation) 
			{
				System.out.println("Match: "+currentSample.equalsIgnoreCase(previousSampleName));
				if(lockFormula) 
				{
					xmlResult.addDerivedParameterFormula("Calibrated_"+selectedParameter, calibrationFormula);
					DateTimeFormatter dtf = DateTimeFormatter.ofPattern("yyyy/MM/dd HH:mm:ss");
					LocalDateTime now = LocalDateTime.now();
					System.out.println(dtf.format(now)); 
					formulaCreatedDate = dtf.format(now);
					runAgain = false;
				}
				else 
					if (!lockFormula) 
					{
						Sample sample = FJPluginHelper.getSample(fcmlQueryElement);

						SElement elem = new SElement(FEML.DerivedParameter);
						elem.setString(FEML.name, "Calibrated_"+selectedParameter);
						elem.setString(FJML.type, FEML.ALGEBRAIC_NAME);
						elem.setString(FJML.formula, calibrationFormula);
						DerivedParameter param = new AlgebraicParameter(FJML.EventNumberDP, elem);
						param.setAttributes(elem, true);
						param.setTransform(sample.getParameter(selectedParameter).createDefaultTransform());
						elem.addContent(param.toString());

						System.out.println(elem.toString());
						DerivedParameterNode node = new DerivedParameterNode(sample.getWorkspace(), elem);
						sample.getSampleNode().addOrReplaceDerivedParameterNode(node);

						DateTimeFormatter dtf = DateTimeFormatter.ofPattern("yyyy/MM/dd HH:mm:ss");
						LocalDateTime now = LocalDateTime.now();
						System.out.println(dtf.format(now)); 
						formulaCreatedDate = dtf.format(now);
					}
				JOptionPane.showMessageDialog(null, "Linear regression applied to "+selectedParameter);
				xmlResult.setWorkspaceString("Calibrated "+selectedParameter);
				pluginState = calibrationState.propagateDrp;
				runAgain = false;
				forceCalculation = false;
				previousSampleName = FJPluginHelper.getSampleName(fcmlQueryElement);
				currentSample = previousSampleName;
				return xmlResult;
			}
		}else
			if( !runAgain  && currentSample.equalsIgnoreCase(previousSampleName) ) 
			{
				System.out.println("Same sample, return false.");
				System.out.println("Run Again: "+runAgain);
				System.out.println("Same sample: "+currentSample.equalsIgnoreCase(previousSampleName));
				return xmlResult;
			}
			else
				if( !runAgain && !currentSample.equalsIgnoreCase(previousSampleName) ) 
				{
					boolean retVal = applyToNewSamplePrompt(fcmlQueryElement,  parameterNamesList);
					if(retVal) 
					{

						if(lockFormula) 
						{
							xmlResult.addDerivedParameterFormula("Calibrated_"+selectedParameter, calibrationFormula);
							DateTimeFormatter dtf = DateTimeFormatter.ofPattern("yyyy/MM/dd HH:mm:ss");
							LocalDateTime now = LocalDateTime.now();
							System.out.println(dtf.format(now)); 
							formulaCreatedDate = dtf.format(now);

							runAgain = false;
						}
						else 
							if (!lockFormula) 
							{
								Sample sample = FJPluginHelper.getSample(fcmlQueryElement);

								SElement elem = new SElement(FEML.DerivedParameter);
								elem.setString(FEML.name, "Calibrated_"+selectedParameter);
								elem.setString(FJML.type, FEML.ALGEBRAIC_NAME);
								elem.setString(FJML.formula, calibrationFormula);
								DerivedParameter param = new AlgebraicParameter(FJML.EventNumberDP, elem);
								param.setAttributes(elem, true);
								param.setTransform(sample.getParameter(selectedParameter).createDefaultTransform());
								elem.addContent(param.toString());

								System.out.println(elem.toString());
								DerivedParameterNode node = new DerivedParameterNode(sample.getWorkspace(), elem);
								sample.getSampleNode().addOrReplaceDerivedParameterNode(node);

								DateTimeFormatter dtf = DateTimeFormatter.ofPattern("yyyy/MM/dd HH:mm:ss");
								LocalDateTime now = LocalDateTime.now();
								System.out.println(dtf.format(now)); //2016/11/16 12:08:43
								formulaCreatedDate = dtf.format(now);
							}
						JOptionPane.showMessageDialog(null, "Linear regression applied to "+selectedParameter);
						xmlResult.setWorkspaceString("Calibrated "+selectedParameter);
						pluginState = calibrationState.propagateDrp;
						previousSampleName = FJPluginHelper.getSampleName(fcmlQueryElement);
						runAgain = false;
						return xmlResult;
					}else if(!retVal) 
					{
						return xmlResult;
					}
				}
		return xmlResult;
	}
}

